import numpy as np
import re
from warnings import warn

from .. import DNPData

import os

_dspfvs_table_10 = {
    2: 44.7500,
    3: 33.5000,
    4: 66.6250,
    6: 59.0833,
    8: 68.5625,
    12: 60.3750,
    16: 69.5313,
    24: 61.0208,
    32: 70.0156,
    48: 61.3438,
    64: 70.2578,
    96: 61.5052,
    128: 70.3789,
    192: 61.5859,
    256: 70.4395,
    384: 61.6263,
    512: 70.4697,
    1024: 70.4849,
    1536: 61.6566,
    2048: 70.4924,
}

_dspfvs_table_11 = {
    2: 46.0000,
    3: 36.5000,
    4: 48.0000,
    6: 50.1667,
    8: 53.2500,
    12: 69.5000,
    16: 72.2500,
    24: 70.1667,
    32: 72.7500,
    48: 70.5000,
    64: 73.0000,
    96: 70.6667,
    128: 72.5000,
    192: 71.3333,
    256: 72.2500,
    384: 71.6667,
    512: 72.1250,
    1024: 72.0625,
    1536: 71.9167,
    2048: 72.0313,
}

_dspfvs_table_12 = {
    2: 46.311,
    3: 36.530,
    4: 47.870,
    6: 50.229,
    8: 53.289,
    12: 69.551,
    16: 71.600,
    24: 70.184,
    32: 72.138,
    48: 70.528,
    64: 72.348,
    96: 70.700,
    128: 72.524,
    192: 71.3333,
    256: 72.2500,
    384: 71.6667,
    512: 72.1250,
    1024: 72.0625,
    1536: 71.9167,
    2048: 72.0313,
}

_dspfvs_table_13 = {
    2: 2.750,
    3: 2.833,
    4: 2.875,
    6: 2.917,
    8: 2.938,
    12: 2.958,
    16: 2.969,
    24: 2.979,
    32: 2.984,
    48: 2.989,
    64: 2.992,
    96: 2.995,
}

_required_params = {
    "acqus": [
        "SW_h",
        "RG",
        "DECIM",
        "DSPFIRM",
        "DSPFVS",
        "BYTORDA",
        "TD",
        "SFO1",
    ],
    "acqu2s": ["TD", "SW_h"],
    "acqu3s": ["TD", "SW_h"],
}


def find_group_delay(attrs_dict):
    """Determine group delay from tables

    Args:
        attrs_dict (dict): dictionary of topspin acquisition parameters

    Returns:
        float: Group delay. Number of points FID is shifted by DSP. The ceiling of this number (group delay rounded up) is the number of points should be removed from the start of the FID.
    """

    # This must be revisited
    group_delay = 0
    if attrs_dict["DSPFVS"] >= 13 and "GRPDLY" in attrs_dict.keys():
        group_delay = attrs_dict["GRPDLY"]
        if group_delay > 0:
            return group_delay

    elif attrs_dict["DECIM"] == 1.0:
        pass
    else:
        if attrs_dict["DSPFVS"] == 10:
            group_delay = _dspfvs_table_10[int(attrs_dict["DECIM"])]
        elif attrs_dict["DSPFVS"] == 11:
            group_delay = _dspfvs_table_11[int(attrs_dict["DECIM"])]
        elif attrs_dict["DSPFVS"] == 12:
            group_delay = _dspfvs_table_12[int(attrs_dict["DECIM"])]
        elif attrs_dict["DSPFVS"] == 13:
            group_delay = _dspfvs_table_13[int(attrs_dict["DECIM"])]
        else:
            print(
                "GRPDLY and DSPFVS parameters not found in acqus file, setting group delay to 0"
            )

    return group_delay


# This function does too much, should be broken into smaller functions
def import_topspin(path, assign_vdlist=False, verbose=False):
    """Import topspin data and return dnpdata object

    Args:
        path (str): Directory of data
        assign_vdlist: False, or the name of dimension to assign topspin vdlist
        verbose (bool): Print additional output for troubleshooting

    Returns:
        dnpdata: topspin data
    """
    dir_list = os.listdir(path)  # All files and folders in directory
    if verbose:
        print("Files in directory:")
        for each in dir_list:
            print(" ", each)

    # Load Acquisition Parameters
    if verbose:
        print("Loading acqus")
    acqus_params = load_acqu(os.path.join(path, "acqus"), verbose=verbose)

    dims = [
        "t2"
    ]  # this may cause issues if the first dimension is not the direct time dimension

    if "fid" in dir_list:
        bin_filename = "fid"
    else:
        bin_filename = "ser"

    if verbose:
        print("Binary File:", bin_filename)

    if acqus_params["BYTORDA"] == 0:
        endian = "<"
    else:
        endian = ">"

    if verbose:
        print("endian", endian)

    topspin_major_version = int(acqus_params["topspin"].split(".")[0])

    # Is this incorrect?
    # Most topspin data I've seen is i4, however, later versions seem to have i8
    # float is also possible
    if topspin_major_version >= 4:
        data_bytes = 8
    else:
        data_bytes = 4

    if acqus_params["DTYPA"] == 0:
        data_type = "i"
    elif acqus_params["DTYPA"] == 2:
        data_type = "f"

    data_type = data_type + str(data_bytes)

    if verbose:
        print("data type:", data_type)

    raw = load_bin(os.path.join(path, bin_filename), dtype=endian + data_type)

    # Is data always complex?
    values = raw[0::2] + 1j * raw[1::2]  # convert to complex

    group_delay = find_group_delay(acqus_params)
    if verbose:
        print("group delay", group_delay)

    group_delay = int(np.floor(group_delay))  # should this be floor or ceil?

    # why is dividing by 2 required?
    t2 = 1.0 / acqus_params["SW_h"] * np.arange(0, int(acqus_params["TD"] / 2))
    if verbose:
        print("points in FID:", acqus_params["TD"] / 2)

    # Handle t2 group delay
    # t2 = t2[slice(group_delay, int(acqus_params["TD"] / 2))] # Alternative method
    t2 = t2[group_delay:]

    coords = [t2]

    # This will not work for vdlist data
    if "acqu2s" in dir_list:
        if verbose:
            print("Loading acqu2s")
        acqu2s_params = load_acqu(os.path.join(path, "acqu2s"), verbose=verbose)
        dims.insert(0, "t1")
        t1 = 1.0 / acqu2s_params["SW_h"] * np.arange(0, int(acqu2s_params["TD"]))
        coords.insert(0, t1)

    # 3d data must be tested
    if "acqu3s" in dir_list:
        if verbose:
            print("Loading acqu3s")
        acqu3s_params = load_acqu(os.path.join(path, "acqu3s"), verbose=verbose)
        dims.insert(0, "t3")
        t3 = 1.0 / acqu3s_params["SW_h"] * np.arange(0, int(acqu3s_params["TD"]))
        coords.insert(0, t3)

    new_shape = [len(coords[ix]) if dims[ix] != "t2" else -1 for ix in range(len(dims))]
    if verbose:
        print("Raw Data Shape:", np.shape(values))
        print("reshaping data to:", new_shape)

    if assign_vdlist:
        if verbose:
            print("Assigning vdlist to %s dim" % assign_vdlist)
        vdlist = topspin_vdlist(path)
        if assign_vdlist in dims:
            index = dims.index(assign_vdlist)
            coords[index] = vdlist

        else:
            raise ValueError("Could not identify dimension to assign vdlist")

    # reshape values
    values = values.reshape(new_shape)

    # Handle group delay
    values = values[..., slice(group_delay, int(acqus_params["TD"] / 2))]

    # create data object
    topspin_data = DNPData(values, dims, coords, attrs=acqus_params)

    # Add NMR Frequency to attrs
    topspin_data.attrs["nmr_frequency"] = acqus_params["SFO1"] * 1e6

    # reorder so that 't2' is first
    topspin_data.reorder(["t2"])

    return topspin_data


# Load topspin should also handle this
def load_pdata(path, verbose=False):
    """Import prospa processed data

    Args:
        path (str): Directory of pdata
        verbose (bool): If true, print output for troubleshooting

    Returns:
        DNPData: Topspin processed data
    """

    # Directory
    dir_list = os.listdir(path)  # All files and folders in directory
    if verbose:
        print("Files in directory:")
        for each in dir_list:
            print(" ", each)

    proc_params = load_acqu(os.path.join(path, "procs"), verbose=verbose)

    if proc_params["BYTORDP"] == 0:
        endian = "<"
    else:
        endian = ">"

    if verbose:
        print("endian", endian)

    real_raw = load_bin(os.path.join(path, "1r"), dtype=endian + "i4")
    imag_raw = load_bin(os.path.join(path, "1i"), dtype=endian + "i4")

    SW = proc_params["SW_p"]
    offset = proc_params["OFFSET"]  # Reference Offset in ppm
    td_eff = proc_params["TDeff"]
    SI = proc_params["SI"]  # What does SI stand for?
    spectrometer_frequency = proc_params["SF"]  # spectrometer frequency in MHz
    phase_0 = proc_params["PHC0"]  # Phase correction, zeroth order phase
    phase_1 = proc_params["PHC1"]  # Phase correction, first order phase

    f2 = (
        -1 * SW * np.linspace(0, 1, num=SI, endpoint=False) / spectrometer_frequency
        + offset
    )

    raw = real_raw + 1j * imag_raw

    data = DNPData(raw, ["f2"], [f2], attrs=proc_params)
    data.attrs["nmr_frequency"] = spectrometer_frequency
    data.attrs["phase_0"] = phase_0
    data.attrs["phase_1"] = phase_1

    return data


# Is this function obsolete?
def load_acqu(path, required_params=None, verbose=False):
    """Import topspin acqu or proc files

    Args:
        path (str): directory of acqu or proc file
        required_params (list): Only return parameters given
        verbose (bool): If true, print output for troubleshooting

    Returns:
        dict: Dictionary of acqusition parameters
    """

    raw_params = load_topspin_jcamp_dx(path, verbose=False)

    if required_params is not None:
        acqus_params = {}
        for key in required_params:
            acqus_params[key] = raw_params[key]
    else:
        acqus_params = raw_params

    return acqus_params


# Legacy
def load_fid_ser(path, dtype="fid", phase_cycle=None):
    """Depreciated. Used import_topspin instead. Import topspin fid or ser file

    Args:
        path (str): Directory of data
        dtype (str): "fid" for 1D, "ser" or "serPhaseCycle" for 2D
        phase_cycle (list): list of phases used for phase cycling (deg, multiples of 90)

    Returns:
        dnpdata: Topspin data
    """

    warn(
        "This function is deprecated, use load_topspin instead",
        DeprecationWarning,
        stacklevel=2,
    )

    if isinstance(phase_cycle, list):
        for indx, x in enumerate(phase_cycle):
            if x in [0, 360, 720, 1080]:
                phase_cycle[indx] = 1
            elif x in [90, 450, 810, 1170]:
                phase_cycle[indx] = 1j
            elif x in [180, 540, 900, 1260]:
                phase_cycle[indx] = -1
            elif x in [270, 630, 990, 1350]:
                phase_cycle[indx] = -1j

    dir_list = os.listdir(path)

    attrs_dict_list = [
        load_acqu(path, x) for x in ["acqus", "acqu2s", "acqu3s"] if x in dir_list
    ]
    attrs_dict = {}
    for a in attrs_dict_list:
        attrs_dict.update(a)

    important_params_dict = {
        "nmr_frequency": attrs_dict["SFO1"] * 1e6,
        "SW_h": attrs_dict["SW_h"],
        "TD": attrs_dict["TD"],
    }

    higher_dim_pars = {
        x: attrs_dict[x]
        for x in ["SW_h_2", "TD_2", "SFO1_2", "SW_h_3", "TD_3", "SFO1_3"]
        if x in attrs_dict.keys()
    }
    important_params_dict.update(higher_dim_pars)

    if attrs_dict["BYTORDA"] == 0:
        endian = "<"
    else:
        endian = ">"

    if dtype == "fid":
        raw = np.fromfile(os.path.join(path, "fid"), dtype=endian + "i4")
    else:
        raw = np.fromfile(os.path.join(path, "ser"), dtype=endian + "i4")

    data = raw[0::2] + 1j * raw[1::2]  # convert to complex

    group_delay = find_group_delay(attrs_dict)
    group_delay = int(np.ceil(group_delay))

    t = 1.0 / attrs_dict["SW_h"] * np.arange(0, int(attrs_dict["TD"] / 2) - group_delay)

    if "vdlist" in dir_list:
        important_params_dict.update({"VDLIST": topspin_vdlist(path)})

    if dtype == "fid":
        data = data[group_delay : int(attrs_dict["TD"] / 2)] / attrs_dict["RG"]
        coords = [t]
        dims = ["t2"]
    elif dtype == "ser":
        ser_data = data.reshape(int(attrs_dict["TD_2"]), -1).T
        ser_data = (
            ser_data[group_delay : int(attrs_dict["TD"] / 2), :] / attrs_dict["RG"]
        )
        coords = [t, range(int(attrs_dict["TD_2"]))]
        dims = ["t2", "t1"]
        if "acqu2s" in dir_list and "acqu3s" not in dir_list:
            if "VDLIST" in important_params_dict.keys():
                if len(important_params_dict["VDLIST"]) == int(attrs_dict["TD_2"]):
                    coords = [t, important_params_dict["VDLIST"]]
            else:
                if isinstance(phase_cycle, list):
                    length1d = int((np.ceil(attrs_dict["TD"] / 256.0) * 256) / 2)
                    ser_data = data.reshape(-1, int(length1d)).T
                    ser_data = ser_data[group_delay : int(attrs_dict["TD"] / 2), :]
                    phs_facs = np.tile(
                        np.array(phase_cycle),
                        int(attrs_dict["TD_2"]) / len(phase_cycle),
                    )
                    for indx in range(int(attrs_dict["TD_2"])):
                        ser_data[:, indx] = phs_facs[indx] * ser_data[:, indx]
                    ser_data = ser_data.sum(axis=1) / attrs_dict["RG"]
                    coords = [t]
                    dims = ["t2"]

        elif "acqu2s" in dir_list and "acqu3s" in dir_list:
            t = (
                1.0
                / attrs_dict["SW_h"]
                * np.arange(
                    0, int(attrs_dict["TD"] / 2 / int(attrs_dict["TD_3"])) - group_delay
                )
            )
            if "VDLIST" in important_params_dict.keys() and len(
                important_params_dict["VDLIST"]
            ) == int(attrs_dict["TD_2"]):
                coords = [
                    t,
                    important_params_dict["VDLIST"],
                    range(int(attrs_dict["TD_3"])),
                ]
            else:
                coords = [
                    t,
                    range(int(attrs_dict["TD_2"])),
                    range(int(attrs_dict["TD_3"])),
                ]
            dims = ["t2", "t1", "t0"]
            ser_data = ser_data.reshape(
                int(attrs_dict["TD"] / 2 / int(attrs_dict["TD_3"])),
                int(attrs_dict["TD_2"]),
                int(attrs_dict["TD_3"]),
            )
        data = ser_data

    return DNPData(data, dims, coords, important_params_dict)


def topspin_vdlist(path):
    """
    Return topspin vdlist

    Args:
        path (str): Directory of data

    Returns:
        numpy.ndarray: vdlist as numpy array
    """
    fullPath = os.path.join(path, "vdlist")

    with open(fullPath, "r") as f:
        raw = f.read()

    lines = raw.rstrip().rsplit()

    unitDict = {
        "n": 1.0e-9,
        "u": 1.0e-6,
        "m": 1.0e-3,
        "k": 1.0e3,
    }
    vdlist = []
    for line in lines:
        if line[-1] in unitDict:
            value = float(line[0:-1]) * unitDict[line[-1]]
            vdlist.append(value)
        else:
            value = float(line)
            vdlist.append(value)

    vdlist = np.array(vdlist)
    return vdlist


# Legacy import ser
def load_ser(path, dtype=">i4"):
    """Depreciated. Use load bin. Import Topspin Ser file

    Args:
        path (str): Directory of data
        dtype (str): data format for import

    returns:
        raw (np.ndarray): Data from ser file
    """
    warn(
        "This function is deprecated, use load_bin instead",
        DeprecationWarning,
        stacklevel=2,
    )

    raw = np.fromfile(os.path.join(path), dtype=dtype)

    return raw


def load_bin(path, dtype=">i4"):
    """Import Topspin Ser file

    Args:
        path (str): Directory of data
        dtype (str): data format for import

    returns:
        raw (np.ndarray): Data from ser file
    """

    raw = np.fromfile(os.path.join(path), dtype=dtype)

    return raw


def load_title(path="1", title_path=os.path.join("pdata", "1"), title_filename="title"):
    """Import Topspin Experiment Title File

    Args:
        path (str): Directory of title
        title_path (str): Path within experiment of title
        title_filename (str): filename of title

    Returns:
        str: Contents of experiment title file
    """

    path_filename = os.path.join(path, title_path, title_filename)

    with open(path_filename, "r") as f:
        rawTitle = f.read()
    title = rawTitle.rstrip()

    return title


def load_topspin_jcamp_dx(path, verbose=False):
    """Return the contents of topspin JCAMP-DX file as dictionary

    Args:
        path (str): Path to file
        verbose (bool): If true, print output for troubleshooting

    Returns:
        dict: Dictionary of JCAMP-DX file parameters
    """

    attrs = {}

    with open(path, "r") as f:
        for line in f:
            if verbose:
                print(line)
            line = line.rstrip()

            if line[0:3] == "##$":
                key, value = tuple(line[3:].split("= ", 1))

                # Test for array
                if value[0] == "(":
                    x = re.findall("\([0-9]+\.\.[0-9]+\)", value)

                    start, end = tuple(x[0][1:-1].split("..", 1))

                    array_size = int(end) + 1

                    same_line_array = value.split(")", 1)[-1]

                    array = []
                    if same_line_array != "":
                        same_line_array = same_line_array.split(" ")

                        try:
                            same_line_array = [
                                float(x) if "." in x else int(x)
                                for x in same_line_array
                            ]
                        except:
                            pass  # Needed in case where "<>" found in some arrays

                        array += same_line_array

                    while len(array) < array_size:
                        array_line = f.readline().rstrip().split(" ")

                        try:
                            array_line = [
                                float(x) if "." in x else int(x) for x in array_line
                            ]
                        except:
                            pass  # Needed in case where "<>" found in some arrays

                        array += array_line

                    array = np.array(array)

                    attrs[key] = array

                elif value[0] == "<":
                    value = value[1:-1]
                    if "." in value:
                        try:
                            value = float(value)
                        except:
                            pass
                    else:
                        try:
                            value = int(value)
                        except:
                            pass

                    attrs[key] = value
                else:
                    if "." in value:
                        try:
                            value = float(value)
                        except:
                            pass
                    else:
                        try:
                            value = int(value)
                        except:
                            pass

                    attrs[key] = value

            elif line[0:2] == "##":
                # Extract Title and TopSpin Version, needed for data type determination
                #                if 'TOPSPIN' in line.upper():
                if "TITLE" in line:
                    version = line.split(" ")[-1]
                    attrs["topspin"] = version
                try:
                    key, value = tuple(line[2:].split("= ", 1))
                except:
                    pass

    return attrs



import unittest
import dnplab as dnp
import os
from numpy.testing import assert_array_equal
import numpy as np


class save_h5_tester(unittest.TestCase):
    def setUp(self):
        self.x = np.r_[0:10]
        self.y = self.x ** 2
        self.data = dnp.DNPData(self.y, ["x"], [self.x])

        self.ws = {"data": self.data}

    def test_h5_save(self):

        dnp.save(
            self.data,
            os.path.join(".", "unittests", "test_save_DNPData.h5"),
            overwrite=True,
        )

        os.remove(os.path.join(".", "unittests", "test_save_DNPData.h5"))

    def test_h5_save_ws(self):

        dnp.save(
            self.ws,
            os.path.join(".", "unittests", "test_save_DNPData_dict.h5"),
            overwrite=True,
        )

        os.remove(os.path.join(".", "unittests", "test_save_DNPData_dict.h5"))


if __name__ == "__main__":
    pass



import numpy as np
from scipy.optimize import curve_fit
from ..core.data import DNPData


def fit(
    f,
    data,
    dim,
    p0,
    sigma=None,
    absolute_sigma=False,
    check_finite=True,
    bounds=(-1 * np.inf, np.inf),
    method=None,
    jac=None,
    **kwargs
):
    """Fitting function for DNPData

    Args:
        f (func): Function used in scipy.curve_fit
        data (DNPData): Data for fit
        dim (str): Dimension to perform fit along
        p0 (tuple): Initial guess for fit
        kwargs: Additional parameters for scipy.curve_fit

    Returns:
        out (dict): Dictionary of fit, fitting parameters, and error
    """

    fit = data.copy()
    fit.unfold(dim)

    ydata = fit.values
    xdata = fit.coords[dim]

    popt_list = []
    perr_list = []
    for ix in range(fit.shape[1]):
        ydata = fit.values[:, ix]
        out = curve_fit(
            f,
            xdata,
            ydata,
            p0=p0,
            sigma=sigma,
            absolute_sigma=absolute_sigma,
            check_finite=check_finite,
            bounds=bounds,
            method=method,
            jac=jac,
            **kwargs
        )
        fit_values = f(xdata, *out[0])
        fit.values[:, ix] = fit_values
        popt = out[0]
        pcov = out[1]
        perr = np.sqrt(np.diag(pcov))
        popt_list.append(popt)
        perr_list.append(perr)

    p_shape = list(fit.attrs["folded_shape"])
    p_shape[0] = len(p0)
    popt_array = np.array(popt_list).T.reshape(p_shape)
    perr_array = np.array(perr_list).T.reshape(p_shape)

    fit.fold()

    pdims = list(fit.dims)
    pdims[0] = "popt"

    pcoords = list(fit.coords)

    pcoords[0] = np.array(range(0, len(p0)))

    popt_data = DNPData(popt_array, pdims, pcoords)
    perr_data = DNPData(perr_array, pdims, pcoords)

    out = {
        "fit": fit,
        "popt": popt_data,
        "err": perr_data,
    }

    return out



import unittest
import dnplab as dnp
import os


class load_wrapper_tester(unittest.TestCase):
    def setUp(self):
        self.topspin_dir = os.path.join(".", "data", "topspin")
        self.prospa_dir = os.path.join(".", "data", "prospa", "toluene_10mM_Tempone")
        self.vnmrj_dir = os.path.join(
            ".", "data", "vnmrj", "10mM_tempol_in_water_array.fid"
        )

        self.list_prospa_dir = [
            "./data/prospa/toluene_10mM_Tempone/1/data.1d",
            "./data/prospa/toluene_10mM_Tempone/2/data.1d",
            "./data/prospa/toluene_10mM_Tempone/3/data.1d",
            "./data/prospa/toluene_10mM_Tempone/4/data.1d",
            "./data/prospa/toluene_10mM_Tempone/5/data.1d",
            "./data/prospa/toluene_10mM_Tempone/6/data.1d",
            "./data/prospa/toluene_10mM_Tempone/7/data.1d",
            "./data/prospa/toluene_10mM_Tempone/8/data.1d",
            "./data/prospa/toluene_10mM_Tempone/9/data.1d",
            "./data/prospa/toluene_10mM_Tempone/10/data.1d",
        ]
        self.list_index = range(10)
        self.dim_name = "test_dim"

    def test_topspin(self):
        data = dnp.load(os.path.join(self.topspin_dir, str(1)), data_type="topspin")
        self.assertEqual(data.dims[0], "t2")
        self.assertEqual(data.values.size, 8148)
        self.assertAlmostEqual(data.attrs["nmr_frequency"], 14831413.270000001)

    def test_prospa(self):
        dnp.load(os.path.join(self.prospa_dir, str(1), "data.csv"), data_type="prospa")

    def test_vnmrj(self):
        dnp.load(self.vnmrj_dir, data_type="vnmrj")

    def test_import_list(self):
        dnp.load(
            self.list_prospa_dir,
            data_type="prospa",
            dim=self.dim_name,
            coord=self.list_index,
        )


if __name__ == "__main__":
    unittest.main()
    pass



"""Modules for importing (experimental) data and exporting data and results"""

from . import prospa
from . import topspin
from . import delta
from . import vnmrj
from . import specman
from . import bes3t
from . import winepr
from . import h5
from . import power
from . import cnsi
from . import vna
from . import tnmr
from . import random



import numpy as np
import os
import re
import time
import datetime
from scipy.io import loadmat


def get_powers(path, power_file, experiment_list):
    """
    Split power readings files into array of power measurements equal in length to number of spectra in dataset

    Args:
        path (str) : Path to base folder containing power file
        power_file (str) : filename, "power" or "t1_powers"
        experiment_list (list) : list of folder numbers of experiments corresponding to power_file

    Returns:
        power_list (list) : list of power readings equal in length to experiment_list
    """

    if power_file == "power.mat" or power_file == "power.csv" or power_file == "power":
        power_file = "power"
    elif (
        power_file == "t1_powers.mat"
        or power_file == "t1_powers.csv"
        or power_file == "t1_powers"
    ):
        power_file = "t1_powers"
    else:
        raise TypeError("power file not recognized")

    if power_file == "power":
        buffer = 2.5
    elif power_file == "t1_powers":
        buffer = 50
    else:
        raise TypeError("invalid filename")

    try:
        expTime = []
        absTime = []
        for exp in experiment_list:
            opened = open(os.path.join(path, str(exp), "audita.txt"))
            lines = opened.readlines()
            absStart = lines[8].split(" ")[2] + " " + lines[8].split(" ")[3]
            splitup = re.findall(r"[\w']+", absStart)
            absStart = datetime.datetime(
                int(splitup[0]),
                int(splitup[1]),
                int(splitup[2]),
                int(splitup[3]),
                int(splitup[4]),
                int(splitup[5]),
                int(splitup[6]),
            )
            absStart = time.mktime(absStart.utctimetuple())
            start = lines[8].split(" ")[3]
            start = start.split(":")
            hour = int(start[0], 10) * 3600
            minute = int(start[1], 10) * 60
            second = int(start[2].split(".")[0], 10)
            start = second + minute + hour
            absStop = lines[6].split("<")[1].split(">")[0].split(" ")
            absStop = absStop[0] + " " + absStop[1]
            splitup = re.findall(r"[\w']+", absStop)
            absStop = datetime.datetime(
                int(splitup[0]),
                int(splitup[1]),
                int(splitup[2]),
                int(splitup[3]),
                int(splitup[4]),
                int(splitup[5]),
                int(splitup[6]),
            )
            absStop = time.mktime(absStop.utctimetuple())
            stop = lines[6].split(" ")[4]
            stop = stop.split(":")
            hour = int(stop[0], 10) * 3600
            minute = int(stop[1], 10) * 60
            second = int(stop[2].split(".")[0], 10)
            stop = second + minute + hour
            expTime.append(stop - start)
            absTime.append((absStart, absStop))

        threshold = 20

        if os.path.isfile(os.path.join(path, power_file + ".mat")):
            print("Extracted powers from " + power_file + ".mat file")
            openfile = loadmat(os.path.join(path, power_file + ".mat"))
            power = openfile.pop("powerlist")
            power = np.array([x for i in power for x in i])
            exptime = openfile.pop("timelist")
            exptime = np.array([x for i in exptime for x in i])
        elif os.path.isfile(os.path.join(path, power_file + ".csv")):
            print("Extracted powers from " + power_file + ".csv file")
            openfile = open(os.path.join(path, power_file + ".csv", "r"))
            lines = openfile.readlines()
            if len(lines) == 1:
                lines = lines[0].split("\r")
            lines.pop(0)
            timeList = []
            powerList = []
            for line in lines:
                exptime, power = line.split("\r")[0].split(",")
                timeList.append(float(exptime))
                powerList.append(float(power))
            exptime = np.array(timeList)
            power = np.array(powerList)

        step = exptime[1] - exptime[0]
        dp = []
        for i in range(len(power) - 1):
            dp.append((power[i + 1] - power[i]) / step)
        dp = abs(np.array(dp))

        timeBreak = []
        for i in range(len(dp)):
            if dp[i] >= threshold:
                timeBreak.append(exptime[i])

        timeBreak.sort()
        absTime.sort(key=lambda tup: tup[0])
        offSet = absTime[-1][1] - timeBreak[-1] + buffer

        power_list = []
        for timeVals in absTime:
            start = int(timeVals[0] - offSet + buffer)
            stop = int(timeVals[1] - offSet - buffer)
            cutPower = []
            for k in range(0, len(exptime) - 1):
                if start <= exptime[k] <= stop:
                    cutPower.append(power[k])
            powers = round(np.average(cutPower), 3)
            power_list.append(float(powers))
    except:
        raise ImportError("Unable to read the power file")

    return power_list



import numpy as np


def remove_background(data, dim="t2", deg=0, regions=None):
    """Remove polynomial background from data

    Args:
        data (DNPData): Data object
        dim (str): Dimension to perform background fit
        deg (int): Polynomial degree
        regions (None, list): Background regions, by default entire region is background corrected. Regions can be specified as a list of tuples [(min, max), ...]

    Returns:
        DNPData: Background corrected data
    """

    proc_parameters = {
        "dim": dim,
        "deg": deg,
        "regions": regions,
    }

    fit = background(data, dim=dim, deg=deg, regions=regions)
    data = data - fit

    proc_attr_name = "remove_backround"
    data.add_proc_attrs(proc_attr_name, proc_parameters)

    return data


def background(data, dim="t2", deg=0, regions=None):
    """Remove background from data

    Args:
        data (DNPData): Data object
        dim (str): Dimension to perform background fit
        deg (int): Polynomial degree
        regions (None, list): Background regions, by default entire region is background corrected. Regions can be specified as a list of tuples [(min, max), ...]

    Returns:
        DNPData: Background fit
    """

    out = data.copy()
    out.unfold(dim)

    coord = out.coords[dim]

    if regions == None:
        fit_points = [True for x in coord]
    else:
        fit_points = [False for x in coord]
        for region in regions:
            fit_points = [
                fit_points[ix] or ((coord[ix] >= region[0]) & (coord[ix] <= region[1]))
                for ix in range(len(coord))
            ]

    for ix in range(out.shape[1]):
        p = np.polyfit(coord[fit_points], out.values[:, ix][fit_points], deg=deg)
        fit = np.polyval(p, coord)
        out.values[:, ix] = fit

    out.fold()

    return out



import numpy as np
from ..core.data import DNPData


def fid(points=1024, snr=100.0):
    """Generate an FID dataset for testing

    Args:
        points (int): Number of points in FID
        snr (float): Signal to noise ratio of FID

    Returns:
        DNPData: FID in DNPData object
    """

    t2 = np.r_[0 : 1 : 1j * points]

    values = (
        np.exp(1j * 2 * np.pi * 100.0 * t2) * np.exp(-1 * t2 / 0.10)
        + np.random.randn(points) / snr
    )

    attrs = {"nmr_frequency": 300e6}

    return DNPData(values, ["t2"], [t2])


def ir(points=(1024,), snr=10):
    """Generate an IR dataset for testing"""
    return NotImplemented


def nd(*args):
    """Generate n-dimensional dataset for testing"""
    return NotImplemented



from warnings import warn

import re

import numpy as np

__all__ = ["fourier_transform", "inverse_fourier_transform"]


def convert_to_ppm(freq_coord):
    return NotImplemented


def rename_ft_dim(dim, old_string, new_string):
    if re.fullmatch("%s[0-9]*" % old_string, dim) is not None:
        dim = dim.replace(old_string, new_string)

    return dim


def fourier_transform(
    data,
    dim="t2",
    zero_fill_factor=1,
    shift=True,
    convert_to_ppm=True,
):
    """Perform Fourier Transform down dim dimension given in proc_parameters

    Args:
        data (DNPData): Data object
        dim (str): Dimension to Fourier Transform, "t2" by default
        zero_fill_factor (int): Increases the number of points in Fourier transformed dimension by this factor with zero filling, 1 by default
        shift (bool): Apply fftshift to the Fourier transformed data, placing zero frequency at center of dimension
        convert_to_ppm (bool): If true, convert Fourier transformed axis to ppm units by using the "nmr_frequency" in attrs

    Returns:
        dnpdata: data object after FT

    Examples:
        Example for constructing a DNPData object and performing Fourier transform

        >>> import numpy as np
        >>> from matplotlib.scipy import *
        >>> import dnplab as dnp
        >>> x = np.r_[0:1:2048j]
        >>> y = np.exp(1j*2*np.pi*300*x) * np.exp(-5*x)
        >>> data = dnp.DNPData(y, ['t2'], [x])
        >>> data.attrs['nmr_frequency'] = 300e6
        >>> data = dnp.fourier_transform(data)
        >>> figure()
        >>> dnp.plot(data)
        >>> xlabel('f2 (ppm)')
        >>> dnp.show()

    .. Note::
        Assumes dt = t[1] - t[0]
    """

    # handle zero_fill_factor
    zero_fill_factor = int(zero_fill_factor)
    if zero_fill_factor <= 0:
        zero_fill_factor = 1

    proc_parameters = {
        "dim": dim,
        "zero_fill_factor": zero_fill_factor,
        "shift": shift,
        "convert_to_ppm": convert_to_ppm,
    }

    index = data.dims.index(dim)

    dt = data.coords[dim][1] - data.coords[dim][0]
    n_pts = zero_fill_factor * len(data.coords[dim])
    f = (1.0 / (n_pts * dt)) * np.r_[0:n_pts]
    if shift == True:
        f -= 1.0 / (2 * dt)

    if convert_to_ppm:
        if "nmr_frequency" not in data.attrs.keys():
            warn(
                "NMR frequency not found in the attrs dictionary, coversion to ppm requires the NMR frequency. See docs."
            )
        else:
            nmr_frequency = data.attrs["nmr_frequency"]
            f /= nmr_frequency / 1.0e6  # updated

    data.values = np.fft.fft(data.values, n=n_pts, axis=index)

    if shift:
        data.values = np.fft.fftshift(data.values, axes=index)

    data.coords[dim] = f

    new_dim = rename_ft_dim(dim, "t", "f")
    data.rename(dim, new_dim)

    proc_attr_name = "fourier_transform"
    data.add_proc_attrs(proc_attr_name, proc_parameters)

    return data


def inverse_fourier_transform(
    data,
    dim="f2",
    zero_fill_factor=1,
    shift=True,
    convert_from_ppm=True,
):
    """Perform Fourier Transform down dim dimension given in proc_parameters

    Args:
        data (DNPData): Data object
        dim (str): Dimension to inverse Fourier transform, "f2" by default
        zero_fill_factor (int): Increases the number of points in inverse Fourier transformed dimension by this factor with zero filling, 1 by default
        shift (bool): Apply fftshift to the inverse Fourier transformed data, placing zero frequency at center of dimension
        convert_from_ppm (bool): If true, convert Fourier transformed axis from ppm units by using the "nmr_frequency" in attrs

    Returns:
        dnpdata: data object after IFT

    .. Note::
        Assumes dt = f[1] - f[0]
    """

    # handle zero_fill_factor
    zero_fill_factor = int(zero_fill_factor)
    if zero_fill_factor <= 0:
        zero_fill_factor = 1

    proc_parameters = {
        "dim": dim,
        "zero_fill_factor": zero_fill_factor,
        "shift": shift,
        "convert_from_ppm": convert_from_ppm,
    }

    index = data.dims.index(dim)

    df = data.coords[dim][1] - data.coords[dim][0]
    if convert_from_ppm:
        if "nmr_frequency" not in data.attrs.keys():
            warn(
                "NMR frequency not found in the attrs dictionary, coversion from ppm requires the NMR frequency. See docs."
            )
        else:
            nmr_frequency = data.attrs["nmr_frequency"]
            df /= 1 / (nmr_frequency / 1.0e6)  # updated

    n_pts = zero_fill_factor * len(data.coords[dim])
    t = (1.0 / (n_pts * df)) * np.r_[0:n_pts]

    if shift:
        data.values = np.fft.fftshift(data.values, axes=index)

    data.values = np.fft.ifft(data.values, n=n_pts, axis=index)
    data.coords[dim] = t

    new_dim = rename_ft_dim(dim, "f", "t")
    data.rename(dim, new_dim)

    proc_attr_name = "inverse_fourier_transform"
    data.add_proc_attrs(proc_attr_name, proc_parameters)

    return data


def zero_fill():
    """Zero fill data, Not Implemented"""
    return NotImplemented



import numpy as np
from . import mr_properties

#######################################
# EPR Properties of Selected Radicals #
#######################################
#
radicalProperties = {}

# When adding new radicals, make sure to add a literature reference for each entry. Keys in dictonary should all be lower case.

# Free electron
# http://physics.nist.gov/constants
radicalProperties["gfree"] = [2.00231930436153, None, [0]]


# TEMPO in Toluene
# Table 1, page 243 in Lebedev et al., Bioactive Spinlabels, Springer Verlag, 1992
radicalProperties["tempo1"] = [[2.00980, 2.00622, 2.00220], "14N", [16.8, 20.5, 95.9]]

# TEMPO in Methanol
# Table 1, page 243 in Lebedev et al., Bioactive Spinlabels, Springer Verlag, 1992
radicalProperties["tempo2"] = [[2.00909, 2.00621, 2.00222], "14N", [20.2, 20.2, 102.1]]

# BDPA in Polystyrene
# Bennati et al., JMR, 1999. (2H couplings were scaled to 1H)
radicalProperties["bdpa"] = [[2.00263, 2.00260, 2.00257], "1H", [50.2, 34.5, 13.0]]


def radical_properties(name):
    """Return properties of different radicals. At the minimum the g value is returned. If available, large hyperfine couplings to a nucleus are returned. Add new properties or new radicals to mrProperties.py

    Args:

    +-----------+---------------------------------------------------------------+
    | arg       |  returns                                                      |
    +===========+===============================================================+
    | "gfree"   | 2.00231930436153                                              |
    +-----------+---------------------------------------------------------------+
    | "tempo1"  | [[2.00980, 2.00622, 2.00220], "14N", [16.8, 20.5, 95.9]]      |
    +-----------+---------------------------------------------------------------+
    | "tempo2"  | [[2.00909, 2.00621, 2.00222], "14N", [20.2, 20.2, 102.1]]     |
    +-----------+---------------------------------------------------------------+
    | "bdpa"    | [[2.00263, 2.00260, 2.00257], "1H", [50.2, 34.5, 13.0]]       |
    +-----------+---------------------------------------------------------------+

    Returns:
        principle g values and hyperfine coupling tensor
    """

    name = name.lower()
    if isinstance(name, str):
        if name in radicalProperties:
            giso = radicalProperties.get(name)[0]
        else:
            print("Radical doesn't exist in dictonary")
            return
    else:
        print("ERROR: String expected")

    return giso


def show_dnp_properties(radical, mwFrequency, dnpNucleus):
    """Calculate DNP Properties

    Currently only implemented for liquid state experiments

    Args:
        radical:        Radical name, see mrProperties.py
        mwFrequency:    Microwave frequency in (Hz)
        dnpNucleus:     Nucleus for DNP-NMR experiments

    Example:
        .. code-block:: python

            dnp.dnpTools.show_dnp_poperties('gfree', 9.45e9, '1H')
    """

    # http://physics.nist.gov/constants
    mub = 9.27400968e-24
    planck = 6.62606957e-34

    # Get radical properties
    glist = radicalProperties.get(radical)[0]
    nucleus = radicalProperties.get(radical)[1]
    Alist = radicalProperties.get(radical)[2]

    # Get g-value
    g = np.array(glist)
    giso = np.sum(g) / g.size

    B0 = mwFrequency * planck / giso / mub

    # Get hyperfine coupling and calculate isotropic value
    A = np.array(Alist)
    AisoMHz = np.sum(A) / A.size

    gmr_e = mr_properties("0e")
    AisoT = AisoMHz / gmr_e / 2 / np.pi

    if nucleus != None:
        nucSpin = mr_properties(nucleus, "spin")
        n = 2 * nucSpin + 1
        ms = np.linspace(-1.0 * nucSpin, nucSpin, int(n))
        B = B0 + ms * AisoT

    else:
        nucSpin = 0
        B = B0

    print("")
    print("Input Parameters: ")
    print("Radical                  : ", radical)
    print("giso                     :  %8.6f" % giso)
    print("Nucleus                  : ", nucleus)
    print("Nuc Spin                 : ", nucSpin)
    print("Aiso               (MHz) :  %4.2f" % AisoMHz)
    print("")
    print("Predicted Field Values for DNP: ")
    m = 1
    for b in B:
        print("Transition: ", m)
        print("B                    (T) :  %6.4f" % b)
        nmr = mr_properties("1H") * b * 10 / 2 / np.pi
        print("NMR Frequency      (MHz) :  %6.3f" % nmr)
        print("")
        m += 1



import unittest
import dnplab as dnp
import os
from numpy.testing import assert_array_equal


class import_topspin_tester(unittest.TestCase):
    def setUp(self):
        self.testdata = os.path.join(".", "data", "topspin")

    def test_import_topspin_exp1_is_fid(self):
        data = dnp.load(os.path.join(self.testdata, str(1)), data_type="topspin")
        self.assertEqual(data.dims[0], "t2")
        self.assertEqual(data.values.size, 8148)
        self.assertAlmostEqual(data.attrs["nmr_frequency"], 14831413.270000001)

    def test_import_topspin_exp5_is_2d_phcyc(self):
        data = dnp.load(os.path.join(self.testdata, str(5)), data_type="topspin")
        self.assertEqual(data.values.shape[0], 11913)
        self.assertEqual(data.dims, ["t2", "t1"])
        self.assertAlmostEqual(data.attrs["nmr_frequency"], 14831413.270000001)

    def test_import_topspin_exp28_is_2d(self):
        data = dnp.load(os.path.join(self.testdata, str(28)), data_type="topspin")
        self.assertEqual(data.values.shape, (7923, 8))
        self.assertEqual(data.dims, ["t2", "t1"])
        self.assertAlmostEqual(data.attrs["nmr_frequency"], 14831413.270000001)

    def test_import_topspin_jcamp_dx(self):
        attrs = dnp.io.topspin.load_topspin_jcamp_dx(
            os.path.join(self.testdata, "1", "acqus")
        )
        self.assertEqual(attrs["DIGTYP"], 9)
        self.assertAlmostEqual(attrs["O1"], 1413.27)
        assert_array_equal(attrs["XGAIN"], [0, 0, 0, 0])
        assert_array_equal(attrs["TPOAL"], [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5])


class prospa_import_tester(unittest.TestCase):
    def setUp(self):
        self.test_data = os.path.join(".", "data", "prospa", "toluene_10mM_Tempone")

    def test_import_prospa_exp_is_1d(self):
        datas = [
            dnp.load(
                os.path.join(self.test_data, "%i" % expNum, "data.csv"),
                data_type="prospa",
            )
            for expNum in [1, 21, 42]
        ]
        for i, data in enumerate(datas):
            self.assertEqual(data.values.shape, (16384,))
            self.assertEqual(data.dims, ["t2"])
            self.assertAlmostEqual(data.attrs["nmr_frequency"], 14244500.0)
        self.assertAlmostEqual(datas[0].values[365], -0.217937 + 0.24907j)
        self.assertAlmostEqual(datas[1].values[365], 0.0400292 - 0.0756107j)
        self.assertAlmostEqual(datas[2].values[365], 1.09858 - 2.57966j)


class vnmrj_import_tester(unittest.TestCase):
    def setUp(self):
        self.test_data2Ds = [
            os.path.join(".", "data", "vnmrj", s)
            for s in ["10mM_tempol_in_water_array.fid"]
        ]
        self.test_data1Ds = [
            os.path.join(".", "data", "vnmrj", s)
            for s in [
                "10mM_tempol_in_water_mw_40dBm.fid",
                "10mM_tempol_in_water_mw_off.fid",
            ]
        ]

    def test_import_vnmrj_1d(self):
        datas = [dnp.load(path=path, data_type="vnmrj") for path in self.test_data1Ds]
        for i, data in enumerate(datas):
            self.assertEqual(data.values.shape, (131072,))
            self.assertEqual(
                data.dims,
                [
                    "t2",
                ],
            )
            self.assertAlmostEqual(data.attrs["nmr_frequency"], 14244283.4231)
        self.assertAlmostEqual(datas[0].values[365], (-20378767 + 2734659j))
        self.assertAlmostEqual(datas[1].values[365], (-950662 - 138458j))

    def test_import_vnmrj_2d(self):
        datas = [dnp.load(path=path, data_type="vnmrj") for path in self.test_data2Ds]
        for i, data in enumerate(datas):
            self.assertEqual(data.values.shape, (131072, 5))
            self.assertEqual(data.dims, ["t2", "t1"])
            self.assertAlmostEqual(data.attrs["nmr_frequency"], 14244283.4231)
        self.assertAlmostEqual(datas[0].values[365, 3], (-1263136 - 1063328.5j))


class specman_import_tester(unittest.TestCase):
    def setUp(self):
        self.test_data_2D = os.path.join(".", "data", "specman", "test_specman2D.exp")
        self.test_data_4D = os.path.join(".", "data", "specman", "test_specman4D.d01")

    def test_import_specman_2D(self):
        data = dnp.load(self.test_data_2D, data_type="specman")
        self.assertEqual(data.dims, ["t2", "t1"])
        self.assertEqual(data.values.shape, (1500, 80))

    def test_import_specman_4D(self):
        data = dnp.load(self.test_data_4D, data_type="specman")
        self.assertEqual(data.dims, ["t2", "t1", "t0", "t"])
        self.assertEqual(data.values.shape, (1500, 40, 5, 3))


class bes3t_import_tester(unittest.TestCase):
    def setUp(self):
        self.test_data_HYSCORE = os.path.join(".", "data", "bes3t", "HYSCORE.DSC")
        self.test_data_DEER = os.path.join(".", "data", "bes3t", "DEER.DSC")
        self.test_data_ESE = os.path.join(".", "data", "bes3t", "2D_ESE.DTA")
        self.test_data_1D = os.path.join(".", "data", "bes3t", "1D_CW.DTA")
        self.test_data_2D = os.path.join(".", "data", "bes3t", "2D_CW.YGF")

    def test_import_bes3t_HYSCORE(self):
        data = dnp.load(self.test_data_HYSCORE, data_type="xepr")
        self.assertEqual(data.dims, ["t2", "t1"])
        self.assertEqual(data.values.shape, (175, 175))
        self.assertEqual(max(data.coords["t2"]), 3520.0)
        self.assertEqual(max(data.coords["t1"]), 3520.0)

    def test_import_bes3t_DEER(self):
        data = dnp.load(self.test_data_DEER, data_type="xepr")
        self.assertEqual(data.dims, ["t2"])
        self.assertEqual(data.values.shape, (504,))
        self.assertEqual(data.attrs["frequency"], 33.85)

    def test_import_bes3t_ESE(self):
        data = dnp.load(self.test_data_ESE, data_type="xepr")
        self.assertEqual(data.dims, ["t2", "t1"])
        self.assertEqual(data.values.shape, (512, 50))
        self.assertEqual(data.attrs["frequency"], 9.296)

    def test_import_bes3t_1D(self):
        data = dnp.load(self.test_data_1D, data_type="xenon")
        self.assertEqual(data.dims, ["B0"])
        self.assertEqual(data.values.shape, (2250,))
        self.assertEqual(data.attrs["frequency"], 9.804448)

    def test_import_bes3t_2D(self):
        data = dnp.load(self.test_data_2D, data_type="xenon")
        self.assertEqual(data.dims, ["B0", "t1"])
        self.assertEqual(data.values.shape, (1600, 100))
        self.assertEqual(data.attrs["frequency"], 9.627213)


class winepr_import_tester(unittest.TestCase):
    def setUp(self):
        self.test_data_ESP = os.path.join(".", "data", "parspc", "ExampleESP.par")
        self.test_data_1D = os.path.join(".", "data", "parspc", "Example1D.spc")
        self.test_data_2D = os.path.join(".", "data", "parspc", "Example2D.spc")

    def test_import_winepr_ESP(self):
        data = dnp.load(self.test_data_ESP, data_type="esp")
        self.assertEqual(data.dims, ["t2"])
        self.assertEqual(data.values.shape, (1024,))
        self.assertEqual(data.attrs["conversion_time"], 81.92)

    def test_import_winepr_1D(self):
        data = dnp.load(self.test_data_1D, data_type="winepr")
        self.assertEqual(data.dims, ["B0"])
        self.assertEqual(data.values.shape, (512,))
        self.assertEqual(data.attrs["temperature"], 294.2)

    def test_import_winepr_2D(self):
        data = dnp.load(self.test_data_2D, data_type="winepr")
        self.assertEqual(data.dims, ["B0", "t1"])
        self.assertEqual(data.values.shape, (1024, 15))
        self.assertEqual(data.attrs["frequency"], 9.79)


class delta_import_tester(unittest.TestCase):
    def setUp(self):
        self.test_data_1D = os.path.join(".", "data", "delta", "50percCHCL3.jdf")
        self.test_data_2D = os.path.join(".", "data", "delta", "lineshape_drift.jdf")

    def test_import_delta_1D(self):
        data = dnp.load(self.test_data_1D, data_type="delta")
        self.assertEqual(data.dims, ["t2"])
        self.assertEqual(data.values.shape, (16384,))
        self.assertEqual(max(data.coords["t2"]), 0.262128)

    def test_import_delta_2D(self):
        data = dnp.load(self.test_data_2D, data_type="delta")
        self.assertEqual(data.dims, ["t2", "t1"])
        self.assertEqual(data.values.shape, (8192, 256))
        self.assertEqual(max(data.coords["t2"]), 0.5451929600000001)
        self.assertEqual(max(data.coords["t1"]), 11.953125)


if __name__ == "__main__":
    unittest.main()
    pass



import numpy as np
import os
from .. import DNPData
import warnings

__all__ = ["import_winepr", "load_par", "load_spc"]

rename_dict = {
    "MF": "frequency",
    "MP": "power",
    "MPD": "attenuation",
    "HCF": "center_field",
    "RCT": "conversion_time",
    "RTC": "time_constant",
    "GST": "sweep_start",
    "GSI": "sweep_extent",
    "RMA": "modulation_amplitude",
    "RRG": "receiver_gain",
    "JSD": "nscans",
    "TE": "temperature",
    "JUN": "x_unit",
    "SSX": "x_points",
    "HSW": "x_width",
    "XXUN": "x_unit",
    "XXLB": "x_min",
    "XYUN": "y_unit",
    "SSY": "y_points",
    "XYWI": "y_width",
    "XYLB": "y_min",
}

float_params = [
    "frequency",
    "power",
    "center_field",
    "conversion_time",
    "time_constant",
    "sweep_start",
    "sweep_extent",
    "modulation_amplitude",
    "receiver_gain",
    "temperature",
    "x_width",
    "x_min",
    "y_min",
    "y_width",
]

int_params = [
    "attenuation",
    "nscans",
    "x_points",
    "y_points",
]


def import_winepr(path):
    """Import Bruker par/spc data and return DNPData object

    Args:
        path (str) : Path to either .par or .spc file

    Returns:
        parspc_data (object) : DNPData object containing Bruker par/spc data
    """

    pathexten = os.path.splitext(path)[1]
    path = os.path.splitext(path)[0]
    if pathexten == ".par" or pathexten == ".spc":
        path_par = path + ".par"
        path_spc = path + ".spc"

    else:
        raise TypeError("data file must be .spc or .par")

    params = load_par(path_par)
    values, dims, coords, attrs = load_spc(path_spc, params)

    parspc_data = DNPData(values, dims, coords, attrs)

    return parspc_data


def load_par(path):
    """Import contents of .par file

    Args:
        path (str) : Path to .par file

    Returns:
        params (dict) : dictionary of parameters
    """

    params = {}
    with open(path, "r") as f:

        for line in f:
            line = line.rstrip()
            split_line = line.split(" ", 1)

            if len(split_line) == 2:
                key = split_line[0].strip()
                value = split_line[1].strip()
                params[key] = value

    for key in rename_dict:
        if key in params:
            new_key = rename_dict[key]
            if new_key in float_params:
                params[new_key] = float(params[key])
            elif new_key in int_params:
                params[new_key] = int(float(params[key]))
            else:
                params[new_key] = params[key]

    if "DOS" in params:
        params["endian"] = "LIT"
        params["data_type"] = "float32"
    else:
        params["endian"] = "BIG"
        params["data_type"] = "int32"

    return params


def load_spc(path, params):
    """Import data and axes of .spc file

    Args:
        path (str) : Path to .spc file

    Returns:
        abscissa (ndarray) : coordinates for spectrum or spectra
        spec (ndarray) : data values
        params (dict) : updated dictionary of parameters
        dims (list) : dimension labels
    """

    data_format = np.dtype(params["data_type"]).newbyteorder(params["endian"])
    file_opened = open(path, "rb")
    file_bytes = file_opened.read()
    spec = np.frombuffer(file_bytes, dtype=data_format)

    params.pop("data_type", None)
    params.pop("endian", None)

    if "x_points" not in params.keys():
        if "y_points" not in params.keys():
            params["x_points"] = int(len(spec))
        else:
            params["x_points"] = int(len(spec) / params["y_points"])

    if "center_field" not in params.keys() or "x_width" not in params.keys():
        if "sweep_start" in params.keys() and "sweep_extent" in params.keys():
            abscissa = [
                np.linspace(
                    params["sweep_start"],
                    params["sweep_extent"],
                    params["x_points"],
                )
            ]
        else:
            warnings.warn("not axis information, axis is indexed only")
            abscissa = [range(params["x_points"])]
    elif "center_field" in params.keys() and "x_width" in params.keys():
        abscissa = [
            np.linspace(
                params["center_field"] - params["x_width"] / 2,
                params["center_field"] + params["x_width"] / 2,
                params["x_points"],
            )
        ]
    else:
        warnings.warn("unable to define axis, indexed only")
        abscissa = [range(params["x_points"])]

    if "x_unit" in params.keys() and params["x_unit"] in ["G", "T"]:
        if params["x_unit"] == "G":
            abscissa = [x / 10 for x in abscissa]
        elif params["x_unit"] == "T":
            abscissa = [x * 1000 for x in abscissa]
        dims = ["B0"]
    else:
        dims = ["t2"]

    if "y_points" in params.keys() and params["y_points"] != 1:
        spec = np.reshape(spec, (params["x_points"], params["y_points"]), order="F")
        dims.append("t1")

        if "y_min" in params.keys() and "y_width" in params.keys():
            abscissa.append(
                np.linspace(
                    params["y_min"],
                    params["y_min"] + params["y_width"],
                    params["y_points"],
                )
            )
        else:
            warnings.warn("unable to define indirect axis")

    file_opened.close()

    return spec, dims, abscissa, params



"""DNPLab - Bringing the Power of Python to DNP-NMR Spectroscopy"""

from .core.data import DNPData
from .core.ufunc import *
from .core.util import *

from .constants import *
from .fitting import *
from .math import *

from .io import *
from .io.save import save
from .io.load import load

from .analysis.hydration import hydration

from .processing import *
from .widgets import *
from .plotting import *
from .reporting import *
from .version import __version__



import numpy as np
import dnplab as dnp


def get_gauss_3d(std_noise=0.0):
    x = np.r_[0:100]
    y = np.r_[0:100]
    z = np.r_[0:100]

    noise = std_noise * np.random.randn(len(x), len(y), len(z))
    gauss = np.exp(-1.0 * (x - 50) ** 2.0 / (10.0 ** 2))
    gauss_3d = (
        gauss.reshape(-1, 1, 1) * gauss.reshape(1, -1, 1) * gauss.reshape(1, 1, -1)
    )
    gauss_3d += noise
    return x, y, z, gauss_3d


def test3d(std_noise=0.0):
    x, y, z, gauss_3d = get_gauss_3d(std_noise)
    test_data = dnp.DNPData(gauss_3d, ["x", "y", "z"], [x, y, z])
    return test_data



import unittest
from numpy.testing import assert_array_equal
import dnplab as dnp
import numpy as np


class dnpNMR_tester(unittest.TestCase):
    def setUp(self):
        pts = 1024
        omega = 50 * np.pi
        tau = 0.1
        t2 = np.r_[0 : 1 : 1j * pts]
        y = np.exp(1j * t2 * omega) * np.exp(-1 * t2 / tau)
        self.data = dnp.DNPData(y, ["t2"], [t2])
        self.data.attrs["nmr_frequency"] = 400e6

    def test_basic_nmr_processing(self):

        data = dnp.remove_background(self.data)

        data = dnp.left_shift(data)

        data = dnp.apodize(data, lw=1)

        data = dnp.fourier_transform(data)

        data = dnp.autophase(data)

    def test_calc_enhancement(self):
        pass


class dnpNMR_tester_sim(unittest.TestCase):
    def setUp(self):
        p1 = np.array([0, 0, 0, 1, 2, 3, 4, 3, 2, 1, 0, 0, 0])
        p2 = np.array([0, 1, 2, 3, 4, 3, 2, 1, 0, 0, 0, 0, 0])
        p3 = np.array([0, 0, 0, 0, 0, 1, 2, 3, 4, 3, 2, 1, 0])
        self.data = dnp.DNPData(
            np.array([p1, p2, p3]).T,
            ["x", "t2"],
            [np.arange(0, len(p1)), np.arange(0, 3)],
        )

    def test_align(self):
        self.aligned_data = dnp.align(self.data, dim="x")
        assert_array_equal(
            self.aligned_data.values,
            np.array(
                [
                    [0, 0, 0, 1, 2, 3, 4, 3, 2, 1, 0, 0, 0],
                    [0, 0, 0, 1, 2, 3, 4, 3, 2, 1, 0, 0, 0],
                    [0, 0, 0, 1, 2, 3, 4, 3, 2, 1, 0, 0, 0],
                ]
            ).T,
        )



import numpy as np
from scipy.special import wofz


def voigtian(x, x0, sigma, gamma, integral=1.0):
    r"""Voigtian distribution. Lineshape given by a convolution of Gaussian and Lorentzian distributions.

    Args:
        x (array_like): input x
        x0 (float): center of distribution
        sigma (float): Gaussian Linewidth. Standard deviation of Gaussian distribution.
        gamma (float): Lorentzian linewidth. 2*gamma is the full width at half maximum (FWHM).
        integral (float): Integral of distribution

    Returns:
        ndarray: Voigtian distribution

    :math:`f(x; x_0, \sigma, \gamma) = \frac{\operatorname{Re}[w(z)]}{\sigma \sqrt{2 \pi}}

    where,

    :math:`z = \frac{x + i\gamma}{\sigma \sqrt{2}}

    """
    z = ((x0 - x) + 1j * gamma) / (sigma * np.sqrt(2.0))
    out = np.real(wofz(z)) / (sigma * np.sqrt(2 * np.pi))
    return integral * out


def gaussian(x, x0, sigma, integral=1.0):
    r"""Gaussian distribution.

    Args:
        x (array_like): input x
        x0 (float): Center of distribution
        sigma (float): Standard deviation of Gaussian distribution
        integral (float): Integral of distribution

    Returns:
        ndarray: Gaussian distribution

    :math:`f(x; x_0, \sigma) = \frac{1}{\sigma \sqrt{2 \pi}} \exp{\left(\frac{(x-x_0)^2}{2 \sigma^2}\right)}
    """
    return (
        integral
        / (sigma * np.sqrt(2 * np.pi))
        * np.exp(-((x - x0) ** 2) / (2 * sigma**2))
    )


def lorentzian(x, x0, gamma, integral=1.0):
    r"""Lorentzian Distribution.

    Args:
        x (array_like): input x
        x0 (float): Center of distribution
        gamma (float): Lorentzian width. 2*gamma is full width at half maximum (FWHM).
        integral (float): Integral of distribution

    Returns:
        ndarray: Lorentzian distribution

    f(x) = \frac{1}{\pi \gamma} \left[\frac{\gamma^2}{(x-x_0)^2 + \gamma^2}\right]
    """
    return (
        integral * (1.0 / (np.pi * gamma)) * gamma**2 / ((x - x0) ** 2 + gamma**2)
    )



import unittest
from numpy.testing import assert_allclose, assert_array_equal
import dnplab as dnp
import numpy as np

# FIX -> Should only use Fit functions, not import or processing functions
class dnpFit_tester(unittest.TestCase):
    def setUp(self):
        self.dim = "x"
        self.x = np.r_[-1:1:100j]
        self.y = dnp.lineshape.gaussian(self.x, 0, 0.1)
        self.p0 = (0, 1)
        self.data = dnp.DNPData(self.y, [self.dim], [self.x])

    def test_fit_function(self):
        out = dnp.fit(dnp.lineshape.gaussian, self.data, dim=self.dim, p0=self.p0)
        fit = out["fit"]
        print(fit)
        print(fit.coords["x"])

        assert_allclose(self.data.values, fit.values)
        assert_array_equal(self.data.coords["x"], fit.coords["x"])



import numpy as np


def _handle_array(x):
    """Handle array or integer input argument for window functions

    Args:
        x (array_like, int): array or integer

    Returns:
        int: length of array or integer input
    """
    if type(x) == int:
        N = x
    else:
        N = len(x)

    return N


def exponential(x, lw):
    """Calculate exponential window function

    Args:
        x (array_like): Vector of points
        lw (int or float): linewidth

    Returns:
        array: exponential window function

    .. math::
        f(x) =  e^{-2t * lw}
    """
    return np.exp(-np.pi * (x - x[0]) * lw)


def gaussian(x, lw):
    r"""Calculate gaussian window function

    Args:
        x (array_like): vector of points
        lw (float): Standard deviation of gaussian window

    Returns:
        array: gaussian window function

    .. math::
        \mathrm{gaussian} = e^{(\sigma * x^{2})}
    """
    sigma = lw / (
        2.0 * np.sqrt(2.0 * np.log(2.0))
    )  # convert FWHM to standard deviation
    return np.exp(-1 * 2.0 * np.pi**2.0 * (x**2.0) * (sigma**2.0))


def hann(x):
    r"""Calculate hann window function

    Args:
        x (array_like): vector of points
        N(int): number of points to return in window function

    Returns:
        ndarray: hann window function

    .. math::
        \mathrm{han} = 0.5 + 0.5\cos(\pi * n / (N-1))
    """
    N = _handle_array(x)

    return 0.5 + 0.5 * np.cos(1.0 * np.pi * np.arange(N) / (N - 1))


def traf(x, lw):
    r"""Calculate traf window function

    Args:
        x (array_like): vector of points
        lw (str): linewidth of traficant window

    Returns:
        ndarray: traf window function

    .. math::
        \mathrm{traf}  &=  (f1 * (f1 + f2)) / (f1^{2} + f2^{2}) &

               f1(t)   &=  \exp(-t * \pi * \mathrm{linewidth[0]}) &

               f2(t)   &=  \exp((t - T) * \pi * \mathrm{linewidth[1]}) &
    """
    T2 = 1.0 / (np.pi * lw)
    t = x
    T = np.max(t)
    E = np.exp(-1 * t / T2)
    e = np.exp(-1 * (T - t) / T2)
    return E * (E + e) / (E**2 + e**2)


def hamming(x):
    r"""Calculate hamming window function

    Args:
        x (array_like): vector of points
        N(int): number of points to return in window function

    Returns:
        ndarray: hamming window function

    .. math::
        \mathrm{hamming} = 0.53836 + 0.46164\cos(\pi * n / (N-1))
    """
    N = _handle_array(x)

    return 0.53836 + 0.46164 * np.cos(1.0 * np.pi * np.arange(N) / (N - 1))


# FIX -> Function does not look correct
def lorentz_gauss(x, exp_lw, gauss_lw, gaussian_max=0):
    r"""Calculate lorentz-gauss window function

    Args:
        x (array_like): vector of points
        N(int): number of points to return in window function
        exp_lw (int or float): exponential linewidth
        gauss_lw (int or float): gaussian linewidth
        gaussian_max (int): location of maximum in gaussian window

    Returns:
        array: gauss_lorentz window function

    .. math::
        \mathrm{lorentz\_gauss} &=  \exp(L -  G^{2}) &

           L(t)    &=  \pi * \mathrm{linewidth[0]} * t &

           G(t)    &=  0.6\pi * \mathrm{linewidth[1]} * (\mathrm{gaussian\_max} * (N - 1) - t) &
    """

    N = len(x)
    expo = np.pi * x * exp_lw
    gaus = 0.6 * np.pi * gauss_lw * (gaussian_max * (N - 1) - x)
    return np.exp(expo - gaus**2).reshape(N)


def sin2(x):
    r"""Calculate sin-squared window function

    Args:
        x (array_like): vector of points
        N(int): number of points to return in window function

    Returns:
        array: sin-squared window function

    .. math::
        \sin^{2}  =  \cos((-0.5\pi * n / (N - 1)) + \pi)^{2}
    """
    N = _handle_array(x)

    return np.cos((-0.5 * np.pi * np.arange(N) / (N - 1)) + np.pi) ** 2



import setuptools
from distutils.core import setup

with open("README.md", "r") as f:
    long_description = f.read()

with open("dnplab/version.py", "r") as f:
    # Define __version__
    exec(f.read())

setup(
    name="dnplab",
    packages=setuptools.find_packages(),
    version=__version__,
    license="MIT",
    description="DNPLab - Bringing the Power of Python to DNP-NMR Spectroscopy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="DNPLab Team",
    url="http://dnpLab.net",
    download_url="",
    project_urls={
        "Documentation": "http://docs.dnpLab.net",
        "Examples": "http://dnplab.net/auto_examples/index.html",
        "Source Code": "https://github.com/DNPLab/dnplab",
        "Web App": "http://odnplab.com/",
    },
    keywords=["ODNP", "DNP", "NMR"],
    python_requires=">=3.6",
    install_requires=[
        "numpy>=1.19",
        "scipy>=1.5",
        "matplotlib>=3.3",
        "h5py>=2.10",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



import numpy as np
from scipy.signal import savgol_filter

from ..core.data import DNPData
from ..processing.integration import integrate

import dnplab as dnp


def calculate_enhancement(integrals, off_spectrum_index=0, return_complex_values=False):
    """Calculate enhancement of a power series. Needs integrals as input

    Args:
        integrals (DNPData):
        off_spectrum_index (int):
        return_complex_values (bool):

    Returns:
        enhancements (DNPData): Enhancement values
    """

    enhancements = integrals.copy()

    if not "experiment_type" in integrals.attrs.keys():

        raise KeyError("Experiment type not defined")

    if integrals.attrs["experiment_type"] != "integrals":

        raise ValueError("dnpdata object does not contain integrals.")

    if integrals.dims[0] == "Power":

        enhancements.attrs["experiment_type"] = "enhancements_P"

        enhancements.values = (
            enhancements.values / enhancements.values[off_spectrum_index]
        )

    elif integrals.dims[0] == "B0":

        enhancements.attrs["experiment_type"] = "enhancements_B0"
        print("This is a DNP enhancement profile. Not implemented yet.")

    else:

        raise TypeError(
            "Integration axis not recognized. First dimension should be Power or B0."
        )

    if return_complex_values == True:
        return enhancements

    elif return_complex_values == False:
        return enhancements.real


def signal_to_noise():
    """Find signal-to-noise ratio

    Returns:
        NotImplemented
    """

    return NotImplemented


def smooth(data, dim="t2", window_length=11, polyorder=3):
    """Apply Savitzky Golay Smoothing

    Args:
        data (DNPData): Data object
        dim (str): Dimension to perform smoothing
        window_length (int): Length of window (number of coefficients)
        polyorder (int): Polynomial order to fit samples

    Returns:
        DNPData: Data with Savitzky Golay smoothing applied
    """
    out = data.copy()

    out.unfold(dim)

    out.values = savgol_filter(out.values, window_length, polyorder, axis=0)

    out.fold()

    return out


def left_shift(data, dim="t2", shift_points=0):
    """Remove points from the left of data

    Args:
        data (dnpdata): Data container for data
        dim (str): Name of dimension to left shift, default is "t2"
        shift_points (int): Number of points to left shift, default is 0.

    Returns:
        dnpdata: data object with left-shifted data
    """

    data = data[dim, shift_points:]

    proc_attr_name = "left_shift"
    proc_parameters = {
        "dim": dim,
        "points": shift_points,
    }
    data.add_proc_attrs(proc_attr_name, proc_parameters)

    return data


def normalize():
    return NotImplemented


def reference(data, dim="f2", old_ref=0, new_ref=0):
    """Function for referencing NMR spectra

    Args:
        data (DNPData): Data for referencing
        dim (str): dimension to perform referencing down. By default this dimension is "f2".
        old_ref (float): Value of old reference
        new_ref (float): New reference value

    Returns:
        DNPData: referenced data
    """

    data = data.copy()

    data.coords[dim] -= old_ref - new_ref

    return data



import numpy as np

import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, Button, RadioButtons


def manual_align(data, dim):
    """Manually align spectra"""

    coord = data.coords[dim]
    max_index = int(data.size / (coord.size**2.0))

    fig, ax = plt.subplots()
    plt.subplots_adjust(left=0.25, bottom=0.25)
    init_index = 0
    delta_index = 1
    l = plt.plot(data.coords["f2"], np.real(data.values))
    ax.margins(x=0)

    axcolor = "lightgoldenrodyellow"
    axindex = plt.axes([0.25, 0.1, 0.65, 0.03], facecolor=axcolor)

    sindex = Slider(
        axindex,
        "index",
        -1 * max_index,
        max_index,
        valinit=init_index,
        valstep=delta_index,
    )

    def update(val):
        index = sindex.val
        ix = 0
        for line in l:
            line.set_ydata(np.roll(data[dim, ix].values.ravel(), index * ix))
            ix += 1
        fig.canvas.draw_idle()

    sindex.on_changed(update)

    reset_ax = plt.axes([0.8, 0.025, 0.1, 0.04])
    reset_button = Button(reset_ax, "Reset", color=axcolor, hovercolor="0.975")

    inc_ax = plt.axes([0.6, 0.025, 0.1, 0.04])
    inc_button = Button(inc_ax, "+", color=axcolor, hovercolor="0.975")

    dec_ax = plt.axes([0.4, 0.025, 0.1, 0.04])
    dec_button = Button(dec_ax, "-", color=axcolor, hovercolor="0.975")

    def reset(event):
        sindex.reset()

    def inc(event):
        sindex.set_val(sindex.val + 1)

    def dec(event):
        sindex.set_val(sindex.val - 1)

    reset_button.on_clicked(reset)
    inc_button.on_clicked(inc)
    dec_button.on_clicked(dec)

    plt.show()
    manual_index = sindex.val

    for ix, x in enumerate(data.coords[dim]):
        data[dim, ix] = np.roll(data[dim, ix].values, manual_index * ix)
        ix += 1

    proc_parameters = {
        "dim": dim,
    }
    proc_attr_name = "manualalign"
    data.add_proc_attrs(proc_attr_name, proc_parameters)

    return data



import os
from . import *

from ..core.util import concat


def load(path, data_type=None, dim=None, coord=None, verbose=False, *args, **kwargs):
    """Import data from different spectrometer formats

    Args:
        path (str, list): Path to data directory or list of directories
        data_type (str): Type of spectrometer data to import (optional). Allowed values: "prospa", "topspin", "delta", "vnmrj", "tnmr", "specman", "xenon", "xepr", "winepr", "esp", "h5", "power", "vna", "cnsi_powers"
        dim (str): If giving directories as list, name of dimension to concatenate data along
        coord (numpy.ndarray): If giving directories as list, coordinates of new dimension
        verbose (bool): If true, print debugging output
        args: Args passed to spectrometer specific import function
        kwargs: Key word args passed to spectrometer specific import function

    Returns:
        data (dnpData): Data object
    """

    if isinstance(path, list):
        if len(coord) != len(path):
            raise ValueError(
                "coord must be a list or array equal in len to the number of paths given"
            )

        data_list = []
        if dim is None:
            dim = "unnamed"
        for filename in path:
            data = load_file(
                filename, data_type=data_type, verbose=verbose, *args, **kwargs
            )
            data_list.append(data)

        return concat(data_list, dim=dim, coord=coord)

    else:
        return load_file(path, data_type=data_type, verbose=verbose, *args, **kwargs)


def load_file(path, data_type=None, verbose=False, *args, **kwargs):
    """Import data from different spectrometer formats

    Args:
        path (str): Path to data directory or file
        data_type (str): Type of spectrometer data to import (optional). Allowed values: "prospa", "topspin", "delta", "vnmrj", "tnmr", "specman", "xenon", "xepr", "winepr", "esp", "h5", "power", "vna", "cnsi_powers"
        verbose (bool): If true, print additional debug outputs
        args: Arguments passed to spectrometer specific import function
        kwargs: Key word arguments passed to spectrometer specific import function

    Returns:
        data (dnpData): Data object
    """

    path = os.path.normpath(path)
    if os.path.isdir(path) and path[-1] != os.sep:
        path = path + os.sep

    if data_type == None:
        data_type = autodetect(path, verbose=verbose)

    if data_type == "prospa":
        return prospa.import_prospa(path, *args, **kwargs)

    elif data_type == "topspin":
        return topspin.import_topspin(path, verbose=verbose, *args, **kwargs)

    elif data_type == "topspin pdata":
        # import_topspin should also handle this type, this is a workaround
        return topspin.load_pdata(path, verbose=verbose, *args, **kwargs)

    elif data_type == "topspin dir":
        return topspin.import_topspin_dir(path, *args, **kwargs)

    elif data_type == "delta":
        return delta.import_delta(path, *args, **kwargs)

    elif data_type == "vnmrj":
        return vnmrj.import_vnmrj(path, *args, **kwargs)

    elif data_type == "tnmr":
        return tnmr.import_tnmr(path, *args, **kwargs)

    elif data_type == "specman":
        return specman.import_specman(path, *args, **kwargs)

    elif data_type in ["xepr", "xenon"]:
        return bes3t.import_bes3t(path, *args, **kwargs)

    elif data_type in ["winepr", "esp"]:
        return winepr.import_winepr(path, *args, **kwargs)

    elif data_type == "h5":
        return h5.load_h5(path, *args, **kwargs)

    elif data_type == "power":
        return power.importPower(path, *args, **kwargs)

    elif data_type == "vna":
        return vna.import_vna(path, *args, **kwargs)

    elif data_type == "cnsi_powers":
        return cnsi.get_powers(path, *args, **kwargs)

    else:
        raise ValueError("Invalid data type: %s" % data_type)


# TODO rename to detect_file_format
def autodetect(test_path, verbose=False):
    """Automatically detect spectrometer format

    Args:
        test_path (str): Test directory
        verbose (bool): If true, print output for debugging

    Returns:
        str: Spectrometer type as string

    """

    if verbose == True:
        print("current directory:", os.getcwd())
        print("data path:", test_path)
        abs_path = os.path.abspath(test_path)
        print("absolute path:", abs_path)

    # Remove trailing separator
    if test_path[-1] == os.sep:
        test_path = test_path[:-1]
        if verbose:
            print("removed trailing separator:", os.sep)

    path_exten = os.path.splitext(test_path)[1]
    if path_exten != "" and verbose:
        print("Extension:", path_exten)

    if path_exten == ".DSC" or path_exten == ".DTA" or path_exten == ".YGF":
        spectrometer_format = "xepr"
    elif path_exten in [".par", ".spc"]:
        spectrometer_format = "winepr"
    elif path_exten in [".d01", ".exp"]:
        spectrometer_format = "specman"
    elif path_exten == ".jdf":
        spectrometer_format = "delta"
    elif (
        os.path.isdir(test_path)
        #        and ("fid" in os.listdir(test_path) or "ser" in os.listdir(test_path))
        and ("acqu" in os.listdir(test_path) or "acqus" in os.listdir(test_path))
    ):
        spectrometer_format = "topspin"
    elif os.path.isdir(test_path) and (
        "proc" in os.listdir(test_path) or "procss" in os.listdir(test_path)
    ):
        spectrometer_format = "topspin pdata"
    elif os.path.isdir(test_path) and path_exten == ".fid":
        spectrometer_format = "vnmrj"
    elif path_exten in [".1d", ".2d", ".3d", ".4d"]:
        spectrometer_format = "prospa"
    elif (
        os.path.isdir(test_path)
        and "acqu.par" in os.listdir(test_path)
        and "data.csv" in os.listdir(test_path)
    ):
        spectrometer_format = "prospa"
    elif path_exten == ".h5":
        spectrometer_format = "h5"
    else:
        raise TypeError(
            "No data type given and autodetect failed to detect format, please specify a format"
        )

    if verbose:
        print("Spectrometer Format:", spectrometer_format)

    return spectrometer_format



from warnings import warn

import numpy as np


def autophase(
    data,
    dim="f2",
    method="search",
    reference_range=None,
    pts_lim=None,
    order="zero",
    pivot=0,
    delta=0,
    phase=None,
    reference_slice=None,
    force_positive=False,
):
    """Automatically phase correct data, or apply manual phase correction

    Args:
        data (DNPData): Data object to autophase
        dim (str): Dimension to autophase
        method (str): Autophase method, "search" by default
        reference_range:
        pts_lim:
        order:
        pivot:
        delta:
        phase:
        reference_slice:
        force_positive:

    Returns:
        DNPData: Autophased data, including attrs "phase0" for order="zero", and "phase1" if order="first"

    .. math::

        \mathrm{data}         &= \exp(-1j * \mathrm{phase}) &

        \mathrm{phase(arctan)} &= \mathrm{arctan}(\mathrm{sum}(\mathrm{data.imag}) / \mathrm{sum}(\mathrm{data.real})) &

        \mathrm{phase(search)} &= \mathrm{argmax}(\mathrm{sum}(phased\_real^{2}) / \mathrm{sum}(phased\_imag^{2})) &

        phased\_real          &= \mathrm{data.real} * \exp(-1j * \mathrm{phase}) &

        phased\_imag          &= \mathrm{data.imag} * \exp(-1j * \mathrm{phase}) &

    """

    data = data.copy()
    if reference_slice == 0:
        raise ValueError(
            "please use indices from 1 to number of slices, i.e. use 1 instead of 0"
        )

    shape_data = np.shape(data.values)
    index = data.dims.index(dim)

    if phase is not None:
        method = "manual"

    if method == "manual":
        if order == "zero" and isinstance(phase, (int, float)):
            data.attrs["phase0"] = phase
        elif order == "zero" and not isinstance(phase, (int, float)):
            raise ValueError(
                "for a zero order phase correction you must supply a single phase"
            )
        elif order == "first" and isinstance(phase, (int, float)):
            data.attrs["phase0"] = phase
            order = "zero"
            warn(
                "method=manual and order=first but only a single phase was given, switching to order=zero"
            )
        elif (
            order == "first"
            and isinstance(phase, (list, np.ndarray))
            and len(phase) == shape_data[index]
        ):
            data.attrs["phase1"] = np.array(phase)
        else:
            raise ValueError(
                "Invalid combination of phase order and phase value(s). Supply float for zero order, array or list for first order"
            )
    else:

        if isinstance(reference_range, (list, tuple)) and len(reference_range) == 2:
            check_data = data[dim, (reference_range[0], reference_range[1])]
        else:
            check_data = data.copy()
            if reference_range is not None:
                warn("reference_range must be None or list/tuple length=2")

        if reference_slice is not None:
            if len(shape_data) == 1:
                reference_slice = None
                temp_data = check_data.values
                warn("ignoring reference_slice, this is 1D data")
            else:
                temp_data = check_data.values[:, reference_slice - 1]
        else:
            temp_data = check_data.values

        if method == "arctan":
            data.attrs["phase0"] = np.arctan(
                np.sum(np.imag(temp_data.reshape(-1, 1)))
                / np.sum(np.real(temp_data.reshape(-1, 1)))
            )
        elif method == "search":
            if pts_lim is not None:
                if len(check_data.coords[dim]) > pts_lim:
                    phasing_x = np.linspace(
                        min(check_data.coords[dim]),
                        max(check_data.coords[dim]),
                        int(pts_lim),
                    ).reshape(-1)
                    if len(check_data.dims) == 1:
                        temp_data = np.interp(
                            phasing_x, check_data.coords[dim], check_data.values
                        ).reshape(-1)
                    else:
                        ind_dim = list(set(data.dims) - set([dim]))[0]
                        ind_shape = data.shape[data.index(ind_dim)]
                        temp_data = np.array(
                            [
                                np.interp(
                                    phasing_x,
                                    check_data.coords[dim],
                                    check_data[dim, :].values[:, indx],
                                ).reshape(-1)
                                for indx in range(ind_shape)
                            ]
                        ).reshape(pts_lim, ind_shape)
            phases_0 = np.linspace(-np.pi / 2, np.pi / 2, 180).reshape(-1)
            rotated_data = (temp_data.reshape(-1, 1)) * np.exp(-1j * phases_0)
            real_imag_ratio = (np.real(rotated_data) ** 2).sum(axis=0) / (
                (np.imag(rotated_data) ** 2).sum(axis=0)
            )
            data.attrs["phase0"] = phases_0[np.argmax(real_imag_ratio)]
        else:
            raise TypeError("Invalid autophase method")

    if order == "zero":
        data.values *= np.exp(-1j * data.attrs["phase0"])
    elif order == "first":
        if method == "manual":
            data.attrs["phase1"] = phase
        else:
            pivot_ratio = pivot / len(data.values)
            data.attrs["phase1"] = np.linspace(
                data.attrs["phase0"] - delta * pivot_ratio,
                data.attrs["phase0"] + delta * (1 - pivot_ratio),
                len(data.values),
            )
        data.values.T.dot(np.exp(-1j * data.attrs["phase1"]))

    else:
        raise TypeError("Invalid order or order & phase pair")

    if force_positive:
        if np.sum(data.values) < 0:
            data.values *= -1

    proc_parameters = {
        "method": method,
        "reference_range": reference_range,
        "reference_slice": reference_slice,
        "force_positive": force_positive,
        "order": order,
        "pivot": pivot,
        "delta": delta,
    }
    proc_attr_name = "autophase"
    data.add_proc_attrs(proc_attr_name, proc_parameters)

    return data


def phase_cycle(data, dim, receiver_phase):
    """Apply phase cycle to data

    Args:
        all_data (dnpdata_collection, dnpdata): data to process
        dim (str): dimension to perform phase cycle
        receiver_phase (numpy.array, list): Receiver Phase 0 (x), 1 (y), 2 (-x), 3 (-y)

    Returns:
        dnpdata: data object after phase cycle applied
    """

    data = data.copy()

    if dim not in data.dims:
        raise ValueError("dim not in dims")

    coord = data.coords[dim]
    receiver_phase = np.array(receiver_phase).ravel()

    proc_parameters = {"dim": dim, "receiver_phase": receiver_phase}

    receiver_phase = np.tile(receiver_phase, int(coord.size / receiver_phase.size))

    index = data.dims.index(dim)

    reshape_size = [1 for k in data.dims]
    reshape_size[index] = len(data.coords[dim])

    data *= np.exp(-1j * (np.pi / 2.0) * receiver_phase.reshape(reshape_size))

    proc_attr_name = "phasecycle"
    data.add_proc_attrs(proc_attr_name, proc_parameters)

    return data


def p0_phase():
    return NotImplemented


def p1_phase():
    return NotImplemented


def phase():
    return NotImplemented



from ..core.data import DNPData
import warnings
import numpy as np
import h5py

None_alias = "__PYTHON_NONE__"  # h5 does not have Null type


python_types = [None]
replace_types = [None_alias]


def load_h5(path):
    """Returns Dictionary of dnpDataObjects

    Args:
        path (str): Path to h5 file

    Returns:
        dnpdata_collection: workspace object with data
    """

    f = h5py.File(path, "r")
    keys_list = f.keys()

    if list(keys_list) == [
        "__DNPDATA__"
    ]:  # If Only DNPData object in h5 file, return DNPData object, not dictionary
        data = read_dnpdata(f["__DNPDATA__"])
        return data

    dnp_dict = {}

    for key in keys_list:

        if f[key].attrs["dnplab_data_type"] == "dnpdata":
            data = read_dnpdata(f[key])
        elif f[key].attrs["dnplab_data_type"] == "dict":
            data = read_dict(f[key])
        else:
            warnings.warn("could not import key: %s" % str(key))

        dnp_dict[key] = data
    return dnp_dict


def read_dnpdata(dnpdata_group):
    coords = []
    dims = []
    attrs = {}
    values = dnpdata_group["values"][:]
    version = dnpdata_group.attrs["dnplab_version"]

    for index in range(len(np.shape(values))):
        dim_key = dnpdata_group["values"].dims[index].keys()[0]  # assumes 1 key only
        coords.append(dnpdata_group["values"].dims[index][dim_key][:])
        dims.append(dim_key)

    for k in dnpdata_group["attrs"].attrs.keys():
        v = dnpdata_group["attrs"].attrs[k]
        if v in replace_types:
            ix = replace_types.index(v)
            v = python_types[ix]
        attrs[k] = v
    for k in dnpdata_group["attrs"]:
        v = dnpdata_group["attrs"][k][:]
        attrs[k] = v

    data = DNPData(values, dims, coords, attrs)

    if "proc_attrs" in dnpdata_group.keys():
        proc_attrs = []
        for k in dnpdata_group["proc_attrs"].keys():
            proc_attrs_name = k.split(":", 1)[1]
            proc_attrs_dict = dict(dnpdata_group["proc_attrs"][k].attrs)
            data.add_proc_attrs(proc_attrs_name, proc_attrs_dict)
    return data


def read_dict(dnpdata_group):
    data = dict(dnpdata_group["attrs"].attrs)
    return data


def save_h5(dataDict, path, overwrite=False):
    """Save workspace in .h5 format

    Args:
        dataDict (dict): dnpdata_collection object to save.
        path (str): Path to save data
        overwrite (bool): If True, h5 file can be overwritten. Otherwise, h5 file cannot be overwritten
    """

    if overwrite:
        mode = "w"
    else:
        mode = "w-"

    keysList = dataDict.keys()

    f = h5py.File(path, mode)

    for key in keysList:
        dnpDataObject = dataDict[key]

        dnpDataGroup = f.create_group(key, track_order=True)
        if isinstance(dnpDataObject, DNPData):
            write_dnpdata(dnpDataGroup, dnpDataObject)
        elif isinstance(dnpDataObject, dict):
            write_dict(dnpDataGroup, dnpDataObject)
        else:
            warnings.warn("Could not write key: %s" % str(key))

    f.close()


def write_dnpdata(dnpDataGroup, dnpDataObject):
    """Takes file/group and writes dnpData object to it

    Args:
        dnpDataGroup: h5 group to save data to
        dnpDataObject: dnpdata object to save in h5 format
    """
    dnpDataGroup.attrs["dnplab_version"] = dnpDataObject.version
    dnpDataGroup.attrs["dnplab_data_type"] = "dnpdata"
    dims_group = dnpDataGroup.create_group("dims")  # dimension names e.g. x,y,z
    attrs_group = dnpDataGroup.create_group("attrs")  # dictionary information
    dnp_dataset = dnpDataGroup.create_dataset("values", data=dnpDataObject.values)

    # Save axes information
    for ix in range(len(dnpDataObject.coords)):
        label = dnpDataObject.dims[ix]
        this_axes = dnpDataObject.coords[ix]
        dims_group.create_dataset(label, data=this_axes)
        dims_group[label].make_scale(label)

        dnp_dataset.dims[ix].attach_scale(dims_group[label])

    # Save Parameters
    for key in dnpDataObject.attrs:
        value = dnpDataObject.attrs[key]

        if isinstance(value, np.ndarray):
            attrs_group.create_dataset(key, data=value)
        else:
            if value in python_types:
                ix = python_types.index(value)
                value = replace_types[ix]
            attrs_group.attrs[key] = value

    # Save proc_steps
    if hasattr(dnpDataObject, "proc_attrs"):
        proc_attrs = dnpDataObject.proc_attrs
        proc_attrs_group = dnpDataGroup.create_group("proc_attrs", track_order=True)
        for ix in range(len(proc_attrs)):
            proc_step_name = proc_attrs[ix][0]
            proc_dict = proc_attrs[ix][1]
            proc_attrs_group_subgroup = proc_attrs_group.create_group(
                "%i:%s" % (ix, proc_step_name)
            )
            for key in proc_dict:
                value = proc_dict[key]
                if value is not None:
                    proc_attrs_group_subgroup.attrs[key] = value


def write_dict(dnpDataGroup, dnpDataObject):
    """Writes dictionary to h5 file

    Args:
        dnpDataGroup (h5py.Group): h5 group to write attrs dictionary
        dnpDataObject (DNPData): DNPData object to write
    """
    #    dnpDataGroup.attrs['dnplab_version'] = dnpDataObject.version
    dnpDataGroup.attrs["dnplab_data_type"] = "dict"
    #    dnpDataGroup.attrs['dnplab_version'] = dnpDataObject.version
    attrs_group = dnpDataGroup.create_group("attrs")

    for key in dnpDataObject.keys():
        attrs_group.attrs[key] = dnpDataObject[key]



import numpy as np

from ..math import window

_windows = {
    "exponential": window.exponential,
    "gaussian": window.gaussian,
    "hann": window.hann,
    "hamming": window.hamming,
    "lorentz_gauss": window.lorentz_gauss,
    "traf": window.traf,
    "sin2": window.sin2,
}


def apodize(data, dim="t2", kind="exponential", **kwargs):
    r"""Apply Apodization to data down given dimension

    Args:
        data (DNPData): Data object
        dim (str): Dimension to apply apodization along, "t2" by default
        kind (str): Type of apodization, "exponential" by default
        kwargs: Arguments to be passed to apodization function

    Returns:
        DNPData: data object with window function applied, including attr "window"

    .. math::

        \mathrm{exponential}    &=  \exp(-2t * \mathrm{linewidth}) &

        \mathrm{gaussian}       &=  \exp((\mathrm{linewidth[0]} * t) - (\mathrm{linewidth[1]} * t^{2})) &

        \mathrm{hamming}        &=  0.53836 + 0.46164\cos(\pi * n/(N-1)) &

        \mathrm{han}            &=  0.5 + 0.5\cos(\pi * n/(N-1)) &

        \mathrm{sin2}           &=  \cos((-0.5\pi * n/(N - 1)) + \pi)^{2} &

        \mathrm{lorentz\_gauss} &=  \exp(L -  G^{2}) &

               L(t)    &=  \pi * \mathrm{linewidth[0]} * t &

               G(t)    &=  0.6\pi * \mathrm{linewidth[1]} * (\mathrm{gaussian\_max} * (N - 1) - t) &

        \mathrm{traf}           &=  (f1 * (f1 + f2)) / (f1^{2} + f2^{2}) &

               f1(t)   &=  \exp(-t * \pi * \mathrm{linewidth[0]}) &

               f2(t)   &=  \exp((t - T) * \pi * \mathrm{linewidth[1]}) &
    """

    data = data.copy()

    coord = data.coords[dim]
    index = data.index(dim)

    window = _windows[kind]
    apwin = window(coord, **kwargs)

    data_shape = data.shape

    new_shape = [1 if ix != index else data_shape[index] for ix in range(data.ndim)]
    apwin = np.reshape(apwin, new_shape)

    data *= apwin

    proc_parameters = {
        "kind": kind,
    }
    for key in kwargs:
        proc_parameters[key] = kwargs[key]
    proc_attr_name = "window"
    data.add_proc_attrs(proc_attr_name, proc_parameters)

    return data



from .data import DNPData
import numpy as np

__all__ = ["generate_data"]


def generate_data(shape):
    size = 1
    dims = []
    coords = []
    for ix, length in enumerate(shape):
        dims.append("x" + str(ix))
        coords.append(np.array(range(length)))
        size *= length

    values = np.random.randn(size)

    return DNPData(values, dims, coords)


def ones(shape, dtype=None):
    values = np.ones(shape, dtype=dtype)
    coords = []
    dims = []
    for ix in range(len(shape)):
        dims.append(str(ix))
        coords.append(np.arange(shape[ix]))

    return DNPData(values, dims, coords)


def ones_like(a):
    return ones(a.shape, a.dtype)


def zeros(shape, dtype=None):
    values = np.zeros(shape, dtype=dtype)
    coords = []
    dims = []
    for ix in range(len(shape)):
        dims.append(str(ix))
        coords.append(np.arange(shape[ix]))

    return DNPData(values, dims, coords)


def zeros_like(a):
    zeros_ = zeros(a.shape, a.dtype)
    zeros_.dims = a.dims
    zeros_.coords = a.coords
    zeros_.attrs = a.attrs
    return zeros(a.shape, a.dtype)


def randn(shape):
    values = np.random.randn(*shape)
    coords = []
    dims = []
    for ix in range(len(shape)):
        dims.append(str(ix))
        coords.append(np.arange(shape[ix]))

    return DNPData(values, dims, coords)


def randn_like(a):
    return randn(a.shape)



import numpy as np
from ..core.data import DNPData
from ..core.util import concat

from scipy.integrate import trapz, cumtrapz


def cumulative_integrate(data, dim="f2", regions=None):
    """Cumulative integration

    Args:
        data (DNPData): Data object
        dim (str): Dimension to perform cumulative integration
        regions (None, list): List of tuples to specify range of integration [(min, max), ...]

    Returns:
        data: cumulative sum of data

    Examples:
        Example showing cumulative integration of lorentzian function

        >>> import numpy as np
        >>> from matplotlib.pylab import *
        >>> import dnplab as dnp
        >>> x = np.r_[-10:10:1000j]
        >>> y = dnp.math.lineshape.lorentzian(x,0,1)
        >>> data = dnp.DNPData(y, ['f2'], [x])
        >>> data_int = dnp.cumulative_integrate(data)
        >>> figure()
        >>> dnp.plot(data)
        >>> dnp.plot(data_int)
        >>> show()


    """

    data = data.copy()

    if regions == None:
        index = data.index(dim)
        data.values = cumtrapz(data.values, data.coords[dim], axis=index, initial=0)
        return data
    else:
        data_list = []
        for region in regions:
            data_list.append(cumulative_integrate(data[dim, region], dim))

        return data_list


def integrate(data, dim="f2", regions=None):
    """Integrate data down given dimension

    Args:
        data (DNPData): Data object
        dim (str): Dimension to perform integration
        regions (None, list): List of tuples, by default entire dimension is integrated

    Returns:
        data: integrals of data
    """

    data = data.copy()
    data.attrs["experiment_type"] = "integrals"

    index = data.index(dim)
    if regions == None:
        data.values = trapz(data.values, data.coords[dim], axis=index)
        data.coords.pop(dim)

    else:
        data_list = []
        for region in regions:
            data_list.append(integrate(data[dim, region], dim))

        x = np.array(list(range(len(data_list))))
        dim_name = "integrals"
        data = concat(data_list, dim_name, coord=x)

    return data
